"""YouTube DownLoad."""

#  playlists
# …
# entries

#  playlists / entries
# title
# id

#  playlist
# entries

#  playlist / entries
# …

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from yt_dlp import YoutubeDL

from rwx import Object
from rwx.log import stream as log

URL = "https://youtube.com"


class Channel(Object):
    def __init__(self, d: dict) -> None:
        self.identifier = d["channel_id"]
        self.title = d["channel"]
        self.followers = int(d["channel_follower_count"])
        self.description = d["description"]
        self.tags = d["tags"]
        # TODO thumbnails
        self.uploader_id = d["uploader_id"]
        self.uploader = d["uploader"]
        self.url = d["channel_url"]
        # TODO entries


class Tab(Object, ABC):
    """YouTube Tab."""

    URL_ROOT = "https://youtube.com"

    def __init__(self, object_id: str) -> None:
        """Set object id.

        :param object_id: object identifier
        :type object_id: str
        """
        self.object_id = object_id
        log.info(self.object_id)
        self.url = self.get_url()
        log.info(self.url)

    def extract(self) -> dict[str, Any]:
        """Return extracted dict.

        :rtype: dict
        """
        yt_dl = Tab.yt_dl(self.get_options())
        return yt_dl.extract_info(self.url, download=False)

    def get_options(self) -> dict:
        """Return options for the action.

        :rtype: dict
        """
        return {
            "extract_flat": True,
            "skip_download": True,
        }

    @abstractmethod
    def get_url(self) -> str:
        """Return URL to access for object.

        :rtype: str
        """

    @staticmethod
    def yt_dl(opt: dict) -> YoutubeDL:
        options = {
            **opt,
            "ignoreerrors": False,
            "quiet": False,
        }
        log.info(options)
        return YoutubeDL(options)


class Playlist(Tab):
    def __init__(self, playlist_id: str) -> None:
        super().__init__(playlist_id)

    def get_url(self) -> str:
        return f"{Tab.URL_ROOT}/playlist?list={self.object_id}"


class Playlists(Tab):
    def __init__(self, channel_id: str) -> None:
        super().__init__(channel_id)

    def get_url(self) -> str:
        return f"{Tab.URL_ROOT}/channel/{self.object_id}/playlists"


class Video(Tab):
    def __init__(self, video_id: str) -> None:
        super().__init__(video_id)
        info = self.extract()
        self.title = info["title"]
        self.fulltitle = info["fulltitle"]
        self.duration = info["duration"]
        self.categories = info["categories"]
        self.tags = info["tags"]
        self.description = info["description"]

    def download(self) -> None:
        Tab.yt_dl(
            {
                "format": "bestvideo[ext=mp4]+bestaudio[ext=mp4]",
                "outtmpl": "%(id)s.%(ext)s",
                "writesubtitles": True,
                "writethumbnail": True,
            },
        ).download([self.url])

    def get_url(self) -> str:
        return f"{Tab.URL_ROOT}/watch?v={self.object_id}"


# channel
# title
# channel_follower_count
# description
# tags
# thumbnails
# uploader_id
# uploader
#  videos / entries
# title
# description truncated
# duration
# thumbnails
# view_count
class Videos(Tab):
    def __init__(self, channel_id: str) -> None:
        super().__init__(channel_id)
        info = self.extract()
        self.title = info["title"]
        self.ids = [v["id"] for v in reversed(info["entries"])]
        self.videos = {}

    def get_url(self) -> str:
        return f"{Tab.URL_ROOT}/channel/{self.object_id}/videos"

    def load(self) -> None:
        done = 0
        for video_id in self.ids:
            self.videos[video_id] = Video(video_id)
            done += 1
            log.info(done)

    def next(self) -> str | None:
        for video_id in self.ids:
            if not Path(f"{video_id}.mp4").exists():
                return video_id
        return None


# ╭──────────╮
# │ download │
# ╰──────────╯


def download_video(video_id: str | None) -> None:
    if video_id:
        ytdl(
            {
                "format": "bestvideo[ext=mp4]+bestaudio[ext=mp4]",
                "outtmpl": "%(id)s.%(ext)s",
                "postprocessors": [
                    {
                        "key": "SponsorBlock",
                        "categories": ["sponsor"],
                    },
                    {
                        "key": "ModifyChapters",
                        "remove_sponsor_segments": ["sponsor"],
                    },
                ],
                "writesubtitles": True,
                "writethumbnail": True,
            },
        ).download([f"{URL}/watch?v={video_id}"])


# ╭─────────╮
# │ extract │
# ╰─────────╯


def extract(opt: dict[str, Any], url: str) -> dict[str, Any]:
    """Return extracted dict.

    :rtype: dict
    """
    return ytdl(
        {
            **opt,
            "extract_flat": True,
            "skip_download": True,
        },
    ).extract_info(url, download=False)


def extract_channel(channel_id: str) -> dict:
    """Return extracted channel dict.

    :param channel_id: channel identifier
    :type channel_id: str
    :rtype: dict
    """
    d = extract({}, f"{URL}/channel/{channel_id}")
    return d


def extract_videos(channel_id: str) -> dict:
    """Return extracted videos dict.

    :param channel_id: channel identifier
    :type channel_id: str
    :rtype: dict
    """
    d = extract({}, f"{URL}/channel/{channel_id}/videos")
    return d


def extract_video(video_id: str) -> dict:
    """Return extracted video dict.

    :param video_id: video identifier
    :type video_id: str
    :rtype: dict
    """
    d = extract({}, f"{URL}/watch?v={video_id}")
    return d


# ╭────────╮
# │ filter │
# ╰────────╯


def filter_video(d: dict) -> dict:
    """Return filtered video dict.

    :param d: video info dict
    :type d: dict
    :rtype: dict
    """
    return {
        "title": d["title"],
    }


def filter_videos(d: dict) -> list[str]:
    """Return filtered videos dict.

    :param d: videos dict
    :type d: dict
    :rtype: dict
    """
    return [entry["id"] for entry in reversed(d["entries"])]


# ╭──────╮
# │ next │
# ╰──────╯


def next_download(videos: list[str]) -> str | None:
    for index, video_id in enumerate(videos):
        if not Path(f"{video_id}.mp4").exists():
            log.info(f"{index} ∕ {len(videos)}")
            return video_id
    return None


# ╭─────╮
# │ url │
# ╰─────╯


def url_channel(channel_id: str) -> str:
    """Return channel URL.

    :param channel_id: channel identifier
    :type channel_id: str
    :rtype: str
    """
    return f"{URL}/channel/{channel_id}"


def url_playlists(channel_id: str) -> str:
    """Return playlists URL.

    :param channel_id: channel identifier
    :type channel_id: str
    :rtype: str
    """
    return f"{url_channel(channel_id)}/playlists"


def url_videos(channel_id: str) -> str:
    """Return videos URL.

    :param channel_id: channel identifier
    :type channel_id: str
    :rtype: str
    """
    return f"{url_channel(channel_id)}/videos"


# ╭──────╮
# │ ytdl │
# ╰──────╯


def ytdl(opt: dict) -> YoutubeDL:
    options = {
        **opt,
        "ignoreerrors": False,
        "quiet": False,
    }
    log.info(options)
    return YoutubeDL(options)
